#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <iostream>
#include <map>
#include "graph.h"
#include <malloc.h>
#include <assert.h>
//#define CG_DEBUG

// Graph
// Sort vertices by node name, dictionary order on names
// Sort edges by node number, dictionary order on pairs of integers

struct spmat spmat_malloc(int nr, int nc, int nnz) {
    struct spmat sp;
    sp.nr = nr;
    sp.nc = nc;
    sp.nnz = nnz;
    sp.IA = (int*)malloc((nr+1)*sizeof(int)); // ind first elt of each row
    sp.JA = (int*)malloc(nnz*sizeof(int)); // final col inds
    sp.INZ = (int*)malloc(nr*sizeof(int)); // nnz of each row
    sp.A = (struct dymatelt*)malloc(nnz*sizeof(struct dymatelt)); // number of events
    int i;
    for(i=0;i<nnz;i++) {
        sp.A[i].n = 0;
        sp.A[i].ne = 0;
        sp.A[i].alloc = 4;
        sp.A[i].start = (float*)malloc(4*sizeof(float));
        sp.A[i].end = (float*)malloc(4*sizeof(float));
    }
    return sp;
}

void spmat_delete(struct spmat sp) {
    free(sp.IA);
    free(sp.JA);
    free(sp.INZ);
    int i;
    for(i=0; i<sp.nnz; i++) free(sp.A[i].start);
    for(i=0; i<sp.nnz; i++) free(sp.A[i].end);
    free(sp.A);
}

struct vertex* vertex_malloc(const char* name, int number) {
    struct vertex* v;
    v = (struct vertex*)malloc(sizeof *v);
    if(v==NULL){mg_error("Error allocating vertex. Not enough memory?");mg_quit(EXIT_FAILURE);}
	
    int len = strlen(name) + 1;
    v->name = (char*)malloc(len * sizeof(char));
    if(v->name==NULL){mg_error("Error allocating vertex name. Not enough memory?");mg_quit(EXIT_FAILURE);}
    strcpy(v->name, name);
    v->number = number;
    return v;
}

int vertex_cmp(struct vertex* v1, struct vertex* v2) {
	char* name1 = v1->name;
	char* name2 = v2->name;
	
	return strcmp(name1, name2); // returns some n>0 if name1 lex comes after name2
}

void vertex_print(struct vertex* v) {
	printf("%s, %d\n", v->name, v->number);
}

void vertex_delete(struct vertex* v) {
	free(v->name);
	free(v);
}

struct edge* edge_malloc(int vertex1, int vertex2, float t_start, float t_end) {
    struct edge* e;
    e = (struct edge*)malloc(sizeof *e);
    if(e==NULL){mg_error("Error allocating edge. Not enough memory?");mg_quit(EXIT_FAILURE);}    
    e->vertex1 = vertex1;
    e->vertex2 = vertex2;
    e->t_start = t_start;
    e->t_end = t_end;
	
    return e;
}

int edge_cmp(struct edge* e1, struct edge* e2) {
    if (e1->vertex1 > e2->vertex1)
        return 1;
    else if (e1->vertex1 == e2->vertex1) {
        if (e1->vertex2 == e2->vertex2)
            return 0;
        else if (e1->vertex2 > e2->vertex2)
            return 1;
    }
    return -1;	
}

void edge_print(int vertex1, int vertex2) {
	printf("(%d, %d)\n", vertex1, vertex2);
}

void edge_delete(struct edge* e) {
	free(e);
}

struct graph* graph_malloc(void) {
    struct graph* g;
    g = (struct graph*)malloc(sizeof *g);
    if(g==NULL){mg_error("Error allocating graph. Not enough memory?");mg_quit(EXIT_FAILURE);}    
	
    g->vertex_set_by_name = NULL;
    g->vertex_set_by_number = NULL;
    g->n_vertices = 0;
    g->edge_set = NULL;
    g->n_edges = 0;
    //    g->sp = NULL;
	
    return g;
}

void graph_vertex_malloc(struct graph* g, int n_vertices) {	
	struct vertex** vertex_set_by_name;
	struct vertex** vertex_set_by_number;
	vertex_set_by_name = (struct vertex**)malloc(n_vertices * (sizeof *vertex_set_by_name));
	vertex_set_by_number = (struct vertex**)malloc(n_vertices * (sizeof *vertex_set_by_name));

    if(vertex_set_by_name==NULL || vertex_set_by_number==NULL){
        mg_error("Error allocating vertex set by name/number. Not enough memory?"); mg_quit(EXIT_FAILURE);}    
		
	g->vertex_set_by_name = vertex_set_by_name;
	g->vertex_set_by_number = vertex_set_by_number;
	g->n_vertices = n_vertices;
}

void graph_vertex_realloc(struct graph* g, int n_vertices) {	
	struct vertex** vertex_set_by_name;
	struct vertex** vertex_set_by_number;
	vertex_set_by_name = (struct vertex**)realloc(g->vertex_set_by_name, n_vertices * (sizeof *vertex_set_by_name));
	vertex_set_by_number = (struct vertex**)realloc(g->vertex_set_by_number, n_vertices * (sizeof *vertex_set_by_name));

    if(vertex_set_by_name==NULL || vertex_set_by_number==NULL){
        mg_error("Error allocating vertex set by name/number. Not enough memory?"); mg_quit(EXIT_FAILURE);}    
		
	g->vertex_set_by_name = vertex_set_by_name;
	g->vertex_set_by_number = vertex_set_by_number;
	g->n_vertices = n_vertices;
}

void graph_edge_malloc(struct graph* g, int n_edges) {
    struct edge** edge_set;
    edge_set = (struct edge**)malloc(n_edges * (sizeof *edge_set));
    if(edge_set==NULL){mg_error("Error allocating edge set. Not enough memory?");mg_quit(EXIT_FAILURE);}    
	
    g->edge_set = edge_set;
    g->n_edges = n_edges;
}

void graph_edge_realloc(struct graph* g, int n_edges) {
    struct edge** edge_set;
    int i;
    for(i=n_edges; i<g->n_edges; i++) edge_delete(g->edge_set[i]);
    edge_set = (struct edge**)realloc(g->edge_set, n_edges * (sizeof *edge_set));
    if(edge_set==NULL){mg_error("Error allocating edge set. Not enough memory?");mg_quit(EXIT_FAILURE);}	
    g->edge_set = edge_set;
}

void graph_edge_delete(struct graph* g) {
    int i;
    for (i = 0; i < g->n_edges; i++)
        edge_delete(g->edge_set[i]);
    free(g->edge_set);
    g->edge_set = NULL;
}

struct graph* graph_read(char* file_name) {
    FILE* input_file = fopen(file_name, "r");
    if (input_file == NULL) {
        mg_error("Couldn't open network file: %s", file_name);
        mg_quit(EXIT_FAILURE);
    }
    
    //	char input_buffer[16];
    //	fgets(input_buffer, sizeof input_buffer, input_file);
    //  rewind(input_file);

    struct graph* g = graph_malloc();    
    //  if(strncmp(input_buffer,"LEDA.GRAPH",10)==0) {
    graph_edge_list_read(g, input_file, file_name);

    fclose(input_file);
    printf("Graph %s read. %d vertices, %d edges.\n", file_name, g->n_vertices, g->n_edges);
    //  graph_print_edges(g);
    return g;
}

void graph_edge_list_read(struct graph* g, FILE *input_file, char* file_name) {
    struct vertex* new_vertex;
    int n_alloc_edges = 1024;
    graph_edge_malloc(g,n_alloc_edges); g->n_edges = 0;

    std::map<std::string,int> verts;
    //std::vector<int> edges1,edges2;
    std::map<std::string,int>::iterator it;

    int ret;
    char input_buffer[1024];
    char str1[512], str2[512];
    float t_start, t_end;
    int num1,num2;
    while(fgets(input_buffer, sizeof input_buffer, input_file)) {
        //        if (input_buffer[0]=='#' || input_buffer[0]=='\0') continue;
        ret = sscanf(input_buffer, " %s ", str1);
        if (ret!=1 || str1[0]=='#' || str1[0]=='%') continue;

        ret = sscanf(input_buffer," %f %f %s %s ", &t_start, &t_end, str1, str2);
        if(ret!=4) {mg_error("Error reading network file: %s, line: %s", file_name, input_buffer); mg_quit(EXIT_FAILURE);}
        if(t_start>t_end) {mg_error("Network file: %s, line: %s, t_start > t_end", file_name, input_buffer); mg_quit(EXIT_FAILURE);}

        it = verts.find(std::string(str1));
        if(it!=verts.end()) num1 = it->second;
        else {
            num1 = g->n_vertices;
            g->n_vertices++;
            verts[std::string(str1)] = num1;
        }
        it = verts.find(std::string(str2));
        if(it!=verts.end()) num2 = it->second;
        else {
            num2 = g->n_vertices;
            g->n_vertices++;
            verts[std::string(str2)] = num2;
        }        
        if (num1 != num2) { g->n_edges++; }
        else { continue; } // ignore self edges
        if (n_alloc_edges < g->n_edges) {
            graph_edge_realloc(g,2*n_alloc_edges);            
            n_alloc_edges *= 2;
        }
        if (num1 < num2) {
            g->edge_set[g->n_edges-1] = edge_malloc(num1,num2,t_start,t_end); }
        else {
            g->edge_set[g->n_edges-1] = edge_malloc(num2,num1,t_start,t_end); }
#ifdef CG_DEBUG
        printf("%% %d %d %f %f %s %s\n", num1, num2, t_start, t_end, str1, str2);
#endif
    }
    if (ferror(input_file)) {
        mg_error("Error reading network file: %s", file_name); mg_quit(EXIT_FAILURE);
    }

    graph_vertex_malloc(g,g->n_vertices);
    for(it=verts.begin();it!=verts.end();++it) {
        new_vertex = vertex_malloc((it->first).c_str(), it->second);
        g->vertex_set_by_number[it->second] = new_vertex;
        g->vertex_set_by_name[it->second] = new_vertex;
    }
    graph_sort_vertices(g);

    g->sp = edge_set_to_spmat(g->edge_set, g->n_vertices, g->n_edges, file_name);
    graph_edge_delete(g);
    g->n_edges = g->sp.IA[g->sp.nr];
}

void graph_print_vertices(struct graph* g) {
	struct vertex** vertex_set_by_name = g->vertex_set_by_name;
	int n_vertices = g->n_vertices;
	
	int i;
	for (i = 0; i < n_vertices; i++)
		vertex_print(vertex_set_by_name[i]);
}

void graph_print_edges(struct graph* g) {
    struct spmat sp = g->sp;
	
    int i,j;
    for (j=0;j<sp.nr;j++) {
        for(i=sp.IA[j]; i<sp.IA[j] + sp.INZ[j]; i++) {
            //w = std::min(v_weights[j],v_weights[spJA[i]]);
            edge_print(j,sp.JA[i]);
        }
    }
}

void graph_print_events(struct graph* g) {
    struct spmat sp = g->sp;
	
    int i,j,k,l;
    for (j=0;j<sp.nr;j++) {
        for(i=sp.IA[j]; i<sp.IA[j] + sp.INZ[j]; i++) {
            k = sp.JA[i];
            //w = std::min(v_weights[j],v_weights[spJA[i]]);
            printf("[%d,%d] = [", j, k);
            for (l=0; l<sp.A[i].n; l++) {
                printf("(%f,%f), ", sp.A[i].start[l], sp.A[i].end[l]);
            }
            printf("]\n");
        }
    }
}

void print_spmat(struct spmat sp) {
    int nr = sp.nr;
    int nc = sp.nc;
    int i,j,k,oldend,t;
    printf("%dx%d matrix with %d entries\n", nr, nc, sp.nnz);
    for(j=0;j<nr;j++) {
        oldend = sp.IA[j] + sp.INZ[j];
        for(k=sp.IA[j]; k<oldend; k++) {
            i = sp.JA[k];
            printf("[%d,%d] = [", i, j);
            for (t=0;t<sp.A[k].n;t++) {
                printf("(%f,%f)",sp.A[k].start[t],sp.A[k].end[t]);
            }
            printf("]\n");
        }
    }
}

void print_spmat_sep(int *spIA, int*spINZ,int*spJA,struct compmatelt *spA,int nk) {
    int i,j,k,oldend,t;
    for(j=0;j<nk;j++) {
        oldend = spIA[j] + spINZ[j];
        for(k=spIA[j]; k<oldend; k++) {
            i = spJA[k];
            struct compmatelt cm = spA[k];
            printf("[%d,%d] = %c[", i, j, cm.lr==0?'L':'R');
            int n = cm.lr==0 ? cm.nl : cm.nr;
            float* tstart = cm.lr==0 ? cm.startl : cm.startr;
            float* tend = cm.lr==0 ? cm.endl : cm.endr;
            for (t=0;t<n;t++) {
                printf("(%.3f,%.3f)",tstart[t],tend[t]);
            }
            printf("]\n");
        }
    }
}

void print_spmat_both(int *spIA, int*spINZ,int*spJA,struct compmatelt *spA,int nk) {
    int i,j,k,oldend,t;
    for(j=0;j<nk;j++) {
        oldend = spIA[j] + spINZ[j];
        for(k=spIA[j]; k<oldend; k++) {
            i = spJA[k];
            struct compmatelt cm = spA[k];
            printf("[%d,%d] = L[", i, j);
            int n = cm.nl;
            float* tstart = cm.startl;
            float* tend = cm.endl;
            for (t=0;t<n;t++) {
                printf("(%.3f,%.3f)",tstart[t],tend[t]);
            }
            printf("] R[");
            n = cm.nr;
            tstart = cm.startr;
            tend = cm.endr;
            for (t=0;t<n;t++) {
                printf("(%.3f,%.3f)",tstart[t],tend[t]);
            }
            printf("]\n");
        }
    }
}


void compmatelts_print(int *spJR, int *spJC, struct compmatelt *spA, int m) {
    for(int k=0; k<m; k++) {
        struct compmatelt cm = spA[k];
        int n = cm.lr==0 ? cm.nl : cm.nr;
        float* tstart = cm.lr==0 ? cm.startl : cm.startr;
        float* tend = cm.lr==0 ? cm.endl : cm.endr;
        int i;
        printf("(%d,%d) = %c[",spJR[k],spJC[k],cm.lr==0?'L':'R');
        for(i=0; i<n; i++) {
            printf("(%.3f,%.3f),",tstart[i],tend[i]);
        }
        printf("]\n");
    }
}

void graph_sort_vertices(struct graph* g) {
    struct vertex** vertex_set_by_name = g->vertex_set_by_name;
	
    int i;
    // Move through interval
    for (i = 1; i < g->n_vertices; i++) {
        int j;
        // Move the next element in
        for (j = i; j > 0; j--) {
            int cmp = vertex_cmp(vertex_set_by_name[j], vertex_set_by_name[j-1]);
            struct vertex* temp;
            if (cmp > 0) {
                temp = vertex_set_by_name[j-1];
                vertex_set_by_name[j-1] = vertex_set_by_name[j];
                vertex_set_by_name[j] = temp;
            }
            // If no swaps made, element is in its place
            else break;
        }
    }
}

void graph_sort_edges(struct graph* g) {
    struct edge** edge_set = g->edge_set;
    
    int i;
    // Move through interval
    for (i = 1; i < g->n_edges; i++) {
        int j;
        // Move the next element in
        for (j = i; j > 0; j--) {
            int cmp = edge_cmp(edge_set[j], edge_set[j-1]);
            struct edge* temp;
            if (cmp == 1) {
                temp = edge_set[j-1];
                edge_set[j-1] = edge_set[j];
                edge_set[j] = temp;
            }
            // If no swaps made, element is in its place
            else break;
        }
    }
}

int graph_find_vertex_number(struct graph* g, char* name) {
	struct vertex** vertex_set_by_name = g->vertex_set_by_name;
	int n_vertices = g->n_vertices;
	
	struct vertex* v = vertex_malloc(name, -1);
	
	int left = 0;
	int right = n_vertices - 1;
	int mid = 0;
	while (right >= left) {
		mid = (left + right) / 2;
		int cmp = vertex_cmp(vertex_set_by_name[mid], v);
		if (cmp > 0)
			left = mid + 1;
		else if (cmp < 0)
			right = mid - 1;
		else
			break;
	}
	vertex_delete(v);

	if (right >= left)
		return vertex_set_by_name[mid]->number;
	else
		return -1;
}

char* graph_find_vertex_name(struct graph* g, int number) {
	struct vertex** vertex_set_by_number = g->vertex_set_by_number;
	
	return vertex_set_by_number[number]->name;
}

int graph_find_edge(struct graph* g, int v1, int v2) {
    struct spmat sp = g->sp;
    int left = sp.IA[v1];
    int right = sp.IA[v1] + sp.INZ[v1] - 1;
    while (right >= left) {
        int mid = (left + right) / 2;
        int del = sp.JA[mid] - v2;
        int cmp = ((del>0)-(del<0)); //edge_cmp(edge_set[mid], e);
        if (cmp == 1)
            left = mid + 1;
        else if (cmp == -1)
            right = mid - 1;
        else
            return mid;
    }
    return -1;
}

int graph_is_edge(struct graph* g, int vertex1, int vertex2) {
	int index = graph_find_edge(g, vertex1, vertex2);
	if (index >= 0)
		return 1;
	
	index = graph_find_edge(g, vertex2, vertex1);
	if (index >= 0)
		return 1;
	
	return 0;
}

void graph_delete(struct graph* g) {
    struct vertex** vertex_set_by_name = g->vertex_set_by_name;
    struct vertex** vertex_set_by_number = g->vertex_set_by_number;
    int n_vertices = g->n_vertices;
    struct edge** edge_set = g->edge_set;
    int n_edges = g->n_edges;
	
    int i;
    for (i = 0; i < n_vertices; i++)
        vertex_delete(vertex_set_by_name[i]);
    free(vertex_set_by_name);
    free(vertex_set_by_number);

    if (edge_set != NULL) {
        for (i = 0; i < n_edges; i++)
            edge_delete(edge_set[i]);
        free(edge_set);
    }
    spmat_delete(g->sp);
    free(g);
}

struct alignment* alignment_calloc(struct graph* g1, struct graph* g2) {
	int m = g1->n_vertices;
	int n = g2->n_vertices;
	if (m > n) {
		printf("alignment_malloc: Size of domain is larger than size of range\n");
		return NULL;
	}
	
	struct alignment* a;
	a = (struct alignment*)malloc(sizeof *a);
    if(a==NULL){mg_error("Error allocating alignment. Not enough memory?");mg_quit(EXIT_FAILURE);}    
	
	struct permutation* p;
	p = permutation_calloc(n);
	
	a->domain = g1;
	a->range = g2;
	a->perm = p;
    //	a->invperm = permutation_calloc(n);
	
	a->n_edges_preserved = -1;
	a->n_edges_induced = -1;
	a->score = -1.0;
	a->is_computed = 0;
	
	return a;
}

void alignment_randomize(struct alignment* a) {
	struct permutation* p = a->perm;
	permutation_randomize(p);
}

int alignment_read(struct alignment* a, char* file_name, int flipgorder) {
    struct graph* g1 = a->domain;
    struct graph* g2 = a->range;
    int m = g1->n_vertices;
    int n = g2->n_vertices;
    a->flipgorder = flipgorder;
	
    FILE* input_file = fopen(file_name, "r");
    if (input_file == NULL) {
        mg_error( "Couldn't open alignment file: %s", file_name);
        mg_quit(EXIT_FAILURE);
    }

    struct permutation* p = a->perm;
    int* sequence = p->sequence;

    int ret;
    int i;
    char name1[1000];
    char name2[1000];
    for (i = 0; i < n; i++)
        sequence[i] = -1;
    for (i = 0; i < m; i++) {
        if (!flipgorder) ret = fscanf(input_file, "%s %s", name1, name2);
        else ret = fscanf(input_file, "%s %s", name2, name1);
        if (ret!=2) {
            mg_error( "Error in alignment file: %s. Maybe there are not enough aligned pairs.", file_name);
            mg_quit(EXIT_FAILURE);
        }
        int vertex1 = graph_find_vertex_number(g1, name1);
        int vertex2 = graph_find_vertex_number(g2, name2);
        if (vertex1<0 || vertex2<0) {
            mg_error( "Error in alignment file: %s. Unknown vertex names: %s or %s.",
                      file_name, name1, name2); mg_quit(EXIT_FAILURE);
        }		
        sequence[vertex1] = vertex2;
    }
	
    // Randomize the rest
    int* mapped_to = (int*)calloc(n, sizeof(int));
    if(mapped_to==NULL){mg_error("Error allocating vertex map. Not enough memory?");mg_quit(EXIT_FAILURE);}
    
    for (i = 0; i < n; i++)
        if (sequence[i] >= 0 && sequence[i] < n)
            mapped_to[sequence[i]] = 1;
    int j = 0;
    for (i = 0; i < n; i++)
        if (mapped_to[i] == 0) {
            mapped_to[j] = i;
            j++;
        }
    struct permutation* shuffle_perm = permutation_calloc(n - m);
    knuth_shuffle(shuffle_perm);
    for (i = 0; i < n - m; i++)
        sequence[m+i] = mapped_to[evaluate(shuffle_perm, i)];
	
    free(mapped_to);
    permutation_delete(shuffle_perm);
	
    printf("Alignment %s read.\n", file_name);
	
    return 0;
}

void alignment_tensor(struct alignment* a3, struct alignment* a1, struct alignment* a2, struct tensor_aux_space* taux) {
	tensor(a3->perm, a1->perm, a2->perm, taux);
}

// template <typename T> int sgn(T val) {
//     return (T(0) < val) - (val < T(0));
// }
int alignment_compare(struct alignment* a1, struct alignment* a2, struct carrier* rel) {
    float del = a2->score - a1->score;
    return ((del>0)-(del<0));
}

// converts set of edges to sparse matrix
// unweighted edges, so values are not used
struct spmat edge_set_to_spmat(struct edge **edge_set, int nk, int nnz, char* filename) {
    struct spmat sp = spmat_malloc(nk,nk,nnz);
    int *spWI = (int*)malloc(nk*sizeof(int)); // for the accum. could be done w/ spINZ
    int i,j,k;
    //float Bstart = (float**)malloc(nnz*sizeof(float*));
    //float Bend = (float**)malloc(nnz*sizeof(float*));
    // create uncompressed spmat
    sp.IA[0] = 0;
    for(i=0;i<nk;i++) sp.INZ[i] = 0;
    for(i=0;i<nnz;i++) sp.INZ[edge_set[i]->vertex1]++;
    for(i=0;i<nk;i++) sp.IA[i+1] = sp.IA[i]+sp.INZ[i]; // spIA = [0;cumsum(spINZ)]
    for(i=0;i<nk;i++) sp.INZ[i] = 0;
    for(i=0;i<nnz;i++) {
        j = sp.IA[edge_set[i]->vertex1] + sp.INZ[edge_set[i]->vertex1]++;
        sp.JA[j] = edge_set[i]->vertex2;
        // btw, if sp.A is different for each i, then we have to reorder it here
        sp.A[j].n = 1;
        sp.A[j].ne = 1;
        sp.A[j].start[0] = edge_set[i]->t_start;
        sp.A[j].end[0] = edge_set[i]->t_end;
    }
    // accumulate duplicates
    for(i=0;i<nk;i++) spWI[i] = -1;
    int start,oldend,count = 0;
    for(j=0;j<nk;j++) {
        start = count;
        oldend = sp.IA[j] + sp.INZ[j];
        for(k=sp.IA[j]; k<oldend; k++) {
            i = sp.JA[k];
            if (spWI[i] >= start) {
                transfer_event_values(sp.A, spWI[i], sp.A[spWI[i]].n, k, NULL, NULL);
                sp.INZ[j]--;
                //spA.nnz--; ??
            } else {
                transfer_event_values(sp.A, count, 0, k, NULL, NULL);
                sp.JA[count] = sp.JA[k];
                spWI[i] = count;
                count++;
            }
        }
        sp.IA[j] = start;
    }
    sp.IA[nk] = count;
    sp.nnz = count;
    free(spWI);

#ifdef CG_DEBUG
//    for(j=0;j<nk;j++) {
//        oldend = sp.IA[j] + sp.INZ[j];
//        for(k=sp.IA[j]; k<oldend; k++) {
//            i = sp.JA[k];
//            printf("(%d,%d,%d) [", j, i, sp.A[k]);
//            for(int l=0; l<sp.A[k].n; l++) printf("(%f,%f) ",sp.A[k].start[l],sp.A[k].end[l]);
//            printf("]");
//        }
//    } printf("\n");
#endif

    float tc,tn;
    for(j=0;j<nk;j++) {
        oldend = sp.IA[j] + sp.INZ[j];
        for(k=sp.IA[j]; k<oldend; k++) {
            //i = sp.JA[k];
            //conserved_duration(sp.A[k].n,sp.A[k].start,sp.A[k].end,&tc,&tn);
            //if (tc>0) {
            //    mg_error("Error: Please remove duplicate or overlapping events from network %s\n", filename);
            //    mg_quit(EXIT_FAILURE);
            //}
            sp.A[k].n = merge_overlapping(sp.A[k].n,sp.A[k].start,sp.A[k].end);
        }
    }                
    
    return sp;
}

void dymatelt_realloc(struct dymatelt *spA, int alloc) {
    float *fptemp1 = (float*)realloc(spA->start, alloc*sizeof(float));
    float *fptemp2 = (float*)realloc(spA->end, alloc*sizeof(float));
    if(fptemp1==NULL||fptemp2==NULL){mg_error("Error allocating vertex. Not enough memory?");mg_quit(EXIT_FAILURE);}
    spA->alloc = alloc;
    spA->start = fptemp1;
    spA->end = fptemp2;
    //printf("UU(%p,%d,%d,%d) ", spA->start, malloc_usable_size(spA->start)/sizeof(float), spA->alloc, spA->n);
}    

void transfer_event_values_sp(struct spmat spran, struct dymatelt *spA, int u, int i) {
    if (spA[u].alloc < spran.A[i].n) {
        //printf("XX(%d,%d,%d,%d,%d,%d) ", malloc_usable_size(spAstart[u]), spAalloc[u], spA[u], malloc_usable_size(spran.Astart[i]), spran.Aalloc[i], spran.A[i]);
        dymatelt_realloc(&spA[u], spran.A[i].alloc);
        //printf("TT(%p,%d,%d,%d,%d,%d,%d)\n", spA[u].start, malloc_usable_size(spA[u].start)/sizeof(float), spA[u].alloc, spA[u].n, malloc_usable_size(spran.A[i].start)/sizeof(float), spran.A[i].alloc, spran.A[i].n);
    }
    int l;
    //if (malloc_usable_size(spA[u].start)/sizeof(float) < spran.A[i].n) {
    //    printf("ERROR\n");
    //    printf("XX(%p,%d,%d,%d,%d,%d,%d,%d,%d)\n", spA[u].start, malloc_usable_size(spA[u].start)/sizeof(float), spA[u].alloc, spA[u].n, malloc_usable_size(spran.A[i].start)/sizeof(float), spran.A[i].alloc, spran.A[i].n,u,i);
    //    exit(-1);
    //}
    //else { printf("XX. ");}
    for (l=0; l<spran.A[i].n; l++) {
        spA[u].start[l] = spran.A[i].start[l];
        spA[u].end[l] = spran.A[i].end[l];
    }
    spA[u].n = spran.A[i].n;
}

// merge sorted arrays (ax,bx) -> sortedx and (ay,by) -> sortedy
// |ax|=|bx|=m & |bx|=|by|=n
// why not just do 2 calls...?
void mergesorted2(float *ax, float *ay, int m, float *bx, float *by, int n,
                 float *sortedx, float *sortedy) {
    int i, j, k;
    j = k = 0;
    for (i = 0; i < m + n;) {
        if (j < m && k < n) {
            if (ax[j] < bx[k]) {
                sortedx[i] = ax[j]; sortedy[i] = ay[j];
                j++;
            }
            else {
                sortedx[i] = bx[k]; sortedy[i] = by[k];
                k++;
            }
            i++;
        }
        else if (j == m) {
            for (; i < m + n;) {
                sortedx[i] = bx[k]; sortedy[i] = by[k];
                k++;
                i++;
            }
        }
        else {
            for (; i < m + n;) {
                sortedx[i] = ax[j]; sortedy[i] = ay[j];
                j++;
                i++;
            }
        }
    }
}

// if clen>0 append events from spA[k] to spA[count]
// if clen=0 overwrite " "
void transfer_event_values(struct dymatelt *spA, int count, int clen, int k, float *tempstart, float *tempend) {
    if (spA[count].alloc < spA[count].n + spA[k].n) {
        //printf("transfer_event_values: HI\n");
        dymatelt_realloc(&spA[count], spA[count].alloc + spA[k].alloc);
        //printf("XX(%d,%d,%d,%d,%d,%d,%d) ", malloc_usable_size(spAstart[count]), spAalloc[count], spA[count], malloc_usable_size(spAstart[k]), spAalloc[k], spA[k], clen);
    }
    if (clen>0) {
        int dotemp=0;
        if (tempstart==NULL||tempend==NULL) {
            dotemp = 1;
            tempstart = (float *)malloc(spA[count].alloc*sizeof(float));
            tempend = (float *)malloc(spA[count].alloc*sizeof(float));
            if(tempstart==NULL||tempend==NULL){mg_error("Error allocating vertex. Not enough memory?");mg_quit(EXIT_FAILURE);}
        }
        mergesorted2(spA[count].start,spA[count].end,clen,spA[k].start,spA[k].end,spA[k].n,tempstart,tempend);
        for (int l=0; l<clen+spA[k].n; l++) {
            spA[count].start[l] = tempstart[l];
            spA[count].end[l] = tempend[l];
        }
        if (dotemp) {
            free(tempstart);
            free(tempend);
        }
    } else {
        for (int l=0; l<spA[k].n; l++) {
            spA[count].start[l] = spA[k].start[l];
            spA[count].end[l] = spA[k].end[l];
        }
    }
    //free(spA[k].start);
    //free(spA[k].end);
    spA[count].n = clen + spA[k].n;
}

void transfer_event_values_comp_l(struct spmat spran, struct compmatelt *spA, int u, int i) {
    spA[u].lr = 0;
    spA[u].nl = spran.A[i].n;
    spA[u].startl = spran.A[i].start;
    spA[u].endl = spran.A[i].end;
}

void transfer_event_values_comp_r(struct spmat spran, struct compmatelt *spA, int u, int i) {
    spA[u].lr = 1;
    spA[u].nr = spran.A[i].n;
    spA[u].startr = spran.A[i].start;
    spA[u].endr = spran.A[i].end;
}

void alignment_composite_graph(struct alignment *a, struct compute_aux_space *caux) {
    int i,j,k,l;
    struct graph *domain = a->domain;
    struct graph *range = a->range;    
    int nl,nk = range->n_vertices;
    int nnz = domain->n_edges + range->n_edges;
    
    struct compmatelt *tempA = caux->tempA; // values
    struct compmatelt *spA = caux->spA; // values
    float *tempstart = caux->tempstart;
    float *tempend = caux->tempend;
    for(i=0;i<nnz;i++) {
        tempA[i].nl = 0;
        tempA[i].nr = 0;
        spA[i].nl = 0;
        spA[i].nr = 0;
    }

    int *spIA = caux->spIA; // ind first elt of each row
    int *spJA = caux->spJA; // col inds
    int *spJR = caux->spJR; // col inds
    int *spJC = caux->spJC; // row inds
    int *spINZ = caux->spINZ; // nnz of each row
    int *spWI = caux->spWI; // for the accum. could be done w/ spINZ
    int *v_weights = caux->v_weights;
    for(i=0;i<nk;i++) v_weights[i] = 1; //init cluster sizes
    // create row and col inds
    int v1,v2,u = 0;
    struct spmat spdom = domain->sp;
    struct spmat spran = range->sp;
    for (v1=0;v1<spran.nr;v1++) {
        for(i=spran.IA[v1]; i<spran.IA[v1] + spran.INZ[v1]; i++) {
            v2 = spran.JA[i];
            if (v1<v2) {
                spJR[u] = v1; spJC[u] = v2;
            }
            else {
                spJR[u] = v2; spJC[u] = v1;
            }
            transfer_event_values_comp_l(spran, tempA, u, i);
            tempA[u].ne = 1;
            u++;
        }
    }
    struct permutation *perm = a->perm;
    nl = domain->n_vertices;
    for(i=0;i<nl;i++) v_weights[evaluate(perm,i)]++;
    for (j=0;j<spdom.nr;j++) {
        for(i=spdom.IA[j]; i<spdom.IA[j] + spdom.INZ[j]; i++) {
            //printf("row: %d, col: %d, u: %d, ",v1,spdom.JA[i],u); permutation_print(perm);            
            v1 = evaluate(perm,j);
            v2 = evaluate(perm,spdom.JA[i]);
            if (v1<v2) {
                spJR[u] = v1; spJC[u] = v2;
            }
            else {
                spJR[u] = v2; spJC[u] = v1;
            }
            transfer_event_values_comp_r(spdom, tempA, u, i);
            tempA[u].ne = 1;
            u++;
        }
    }
    // create uncompressed spmat // reordering
    spIA[0] = 0;
    for(i=0;i<nk;i++) spINZ[i] = 0;
    for(i=0;i<nnz;i++) spINZ[spJC[i]]++;
    for(i=0;i<nk;i++) spIA[i+1] = spIA[i]+spINZ[i]; // spIA = [0;cumsum(spINZ)]
    for(i=0;i<nk;i++) spINZ[i] = 0;
    for(i=0;i<nnz;i++) {
        j = spIA[spJC[i]] + spINZ[spJC[i]]++;
        spJA[j] = spJR[i];
        //printf("xx00 %d %d\n", j, i);
        // since spA is different for each i, we reorder it here
        spA[j] = tempA[i];
    }

#ifdef CG_DEBUG
    printf("perm: "); permutation_print(perm);
    printf("A.nl: "); for(i=0;i<nnz;i++) printf("%d ", spA[i].nl); printf("\n");
    printf("A.nr: "); for(i=0;i<nnz;i++) printf("%d ", spA[i].nr); printf("\n");
    printf("A.ne: "); for(i=0;i<nnz;i++) printf("%d ", spA[i].ne); printf("\n");
    printf("JA: "); for(i=0;i<nnz;i++) printf("%d ", spJA[i]); printf("\n");
    printf("JR: "); for(i=0;i<nnz;i++) printf("%d ", spJR[i]); printf("\n");
    printf("JC: "); for(i=0;i<nnz;i++) printf("%d ", spJC[i]); printf("\n");
    printf("IA: "); for(i=0;i<nk+1;i++) printf("%d ", spIA[i]); printf("\n");
    printf("INZ: "); for(i=0;i<nk;i++) printf("%d ", spINZ[i]); printf("\n");
    printf("v_weights: "); for(i=0;i<nk;i++) printf("%d ", v_weights[i]); printf("\n");
    printf("---\n");
#endif

    // accumulate duplicates
    int wi;
    for(i=0;i<nk;i++) spWI[i] = -1;
    int start,oldend,count = 0;
    for(j=0;j<nk;j++) {
        start = count;
        oldend = spIA[j] + spINZ[j];
        for(k=spIA[j]; k<oldend; k++) {
            i = spJA[k];
            wi = spWI[i];
            if (wi >= start) {
                //transfer_event_values(spA, spWI[i], spA[spWI[i]].n, k, tempstart, tempend);
                //spA[spWI[i]] += spA[k];

                if (spA[wi].lr == 0) {
                    assert(spA[k].lr==1);
                    spA[wi].nr = spA[k].nr;
                    spA[wi].startr = spA[k].startr;
                    spA[wi].endr = spA[k].endr;
                } else {
                    assert(spA[k].lr==0);
                    spA[wi].nl = spA[k].nl;
                    spA[wi].startl = spA[k].startl;
                    spA[wi].endl = spA[k].endl;
                }             
                
                spA[wi].ne += spA[k].ne;                
                spINZ[j]--;
            }
            else {
                // transfer_event_values(spA, count, 0, k, tempstart, tempend);
                spA[count] = spA[k];
                //spA[count].ne = spA[k].ne;
                spJA[count] = spJA[k];
                spWI[i] = count;
                count++;
            }
        }
        spIA[j] = start;
    }
    spIA[nk] = count;

#ifdef CG_DEBUG
    printf("perm: "); permutation_print(perm);
    printf("A.nl: "); for(i=0;i<nnz;i++) printf("%d ", spA[i].nl); printf("\n");
    printf("A.nr: "); for(i=0;i<nnz;i++) printf("%d ", spA[i].nr); printf("\n");
    printf("A.ne: "); for(i=0;i<nnz;i++) printf("%d ", spA[i].ne); printf("\n");
    printf("JA: "); for(i=0;i<nnz;i++) printf("%d ", spJA[i]); printf("\n");
    printf("IA: "); for(i=0;i<nk+1;i++) printf("%d ", spIA[i]); printf("\n");
    printf("INZ: "); for(i=0;i<nk;i++) printf("%d ", spINZ[i]); printf("\n");
    printf("v_weights: "); for(i=0;i<nk;i++) printf("%d ", v_weights[i]); printf("\n");
    printf("--------\n");
#endif

    for(i=0;i<nnz;i++) tempA[i] = spA[i];
}

void insertion_sort2(float *arr1, float *arr2, int n) {
    int c,d;
    float temp;
    for (c = 1 ; c <= n - 1; c++) {
        d = c;
        while (d > 0 && arr1[d] < arr1[d-1]) {
            temp      = arr1[d];
            arr1[d]   = arr1[d-1];
            arr1[d-1] = temp;
            temp      = arr2[d];
            arr2[d]   = arr2[d-1];
            arr2[d-1] = temp;
            d--;
        }
    }
}

int merge_overlapping(int n, float* tstart, float* tend) {
    if (n<=1) { return n; }
    int i = 0, m = 0;
    float a,b,c,d;
    a = tstart[i];
    b = tend[i];
    for (i=1; i<n; i++) {
        c = tstart[i];
        d = tend[i];
        if (b < c) {
            tstart[m] = a;
            tend[m] = b;
            m++;
            a = c;
            b = d;
        } else {
            b = d;
        }
    }
    tstart[m] = a;
    tend[m] = b;
    m++;
    return m;
}

void conserved_duration(int n, float* tstart, float* tend, float *Tc, float *Tn) {
    if (n==0) { *Tn = 0; *Tc = 0; return; }
    if (n==1) { *Tn = tend[0]-tstart[0]; *Tc = 0; return; }
    float tc=0, tn=0;
    //printf("before: "); for(int i=0;i<n;i++) { printf("(%f,%f) ", tstart[i], tend[i]); } printf("\n");
    //insertion_sort2(tstart, tend, n);
    #ifdef CG_DEBUG
    //printf("consdur: "); for(int i=0;i<n;i++) { printf("(%f,%f) ", tstart[i], tend[i]); } printf("\n");
    #endif
    int i = 0;
    float a,b,c,d;
    a = tstart[i];
    b = tend[i];
    for (i=1; i<n; i++) {
        c = tstart[i];
        d = tend[i];
        if (b <= c) {
            tn += b-a;
            a = c;
            b = d;
        } else if (b > d) {
            tn += c-a;
            tc += d-c;
            a = d;
        } else {
            tn += c-a;
            tc += b-c;
            a = b;
            b = d;
        }
    }
    tn += b-a;
    *Tc = tc;
    *Tn = tn;
}

void conserved_duration(int nl, float* tstartl, float* tendl,
                        int nr, float* tstartr, float* tendr,
                        float *Tc, float *Tn) {
    //if (n==0) { *Tn = 0; *Tc = 0; return; }
    //if (n==1) { *Tn = tend[0]-tstart[0]; *Tc = 0; return; }
    float tc=0, tn=0;
    //printf("before: "); for(int i=0;i<n;i++) { printf("(%f,%f) ", tstart[i], tend[i]); } printf("\n");
    //insertion_sort2(tstart, tend, n);
#ifdef CG_DEBUG
    //printf("consdur: "); for(int i=0;i<n;i++) { printf("(%f,%f) ", tstart[i], tend[i]); } printf("\n");
#endif
    int il = 0, ir = 0;
    float a=0,b=0,c,d;

    if (il<nl && ir<nr) {
        if (tstartl[il] <= tstartr[ir]) {
            a = tstartl[il]; b = tendl[il];
            il++;
        } else {
            a = tstartr[ir]; b = tendr[ir];
            ir++;
        }
    }
    
    while (il<nl && ir<nr) {
        if (tstartl[il] <= tstartr[ir]) {
            c = tstartl[il]; d = tendl[il];
            il++;
        } else {
            c = tstartr[ir]; d = tendr[ir];
            ir++;
        }    
        
        if (b <= c) {
            tn += b-a;
            a = c;
            b = d;
        } else if (b > d) {
            tn += c-a;
            tc += d-c;
            a = d;
        } else {
            tn += c-a;
            tc += b-c;
            a = b;
            b = d;
        }
    }

    while (il<nl) {
        c = tstartl[il]; d = tendl[il];
        il++;
        
        if (b <= c) {
            tn += b-a;
            a = c;
            b = d;
        } else if (b > d) {
            tn += c-a;
            tc += d-c;
            a = d;
        } else {
            tn += c-a;
            tc += b-c;
            a = b;
            b = d;
        }
    }

    while (ir<nr) {
        c = tstartr[ir]; d = tendr[ir];
        ir++;
        
        if (b <= c) {
            tn += b-a;
            a = c;
            b = d;
        } else if (b > d) {
            tn += c-a;
            tc += d-c;
            a = d;
        } else {
            tn += c-a;
            tc += b-c;
            a = b;
            b = d;
        }
    }
    
    tn += b-a;
    *Tc = tc;
    *Tn = tn;
}


void alignment_edges_preserved(struct alignment *a, struct compute_aux_space *caux,
                               int *n_edges_preserved, int *n_unique_edges,
                               float *t_duration_conserved, float *t_duration_nonconserved) {
    struct graph *range = a->range;    
    int nk = range->n_vertices;
    struct compmatelt *spA = caux->spA; // values
    //float **spAstart = caux->spAstart; // values
    //float **spAend = caux->spAend; // values
    int *spIA = caux->spIA; // ind first elt of each row
    int *spJA = caux->spJA; // col inds
    int *spINZ = caux->spINZ; // nnz of each row
    int *v_weights = caux->v_weights;
    int i,j;
    // calculate preserved and unique edges
    int cis = 0;
    int unique_edges = 0;
    int r,w;
    float Tn=0,Tc=0,tn,tc;
    for (j=0;j<nk;j++) {
        for(i=spIA[j]; i<spIA[j] + spINZ[j]; i++) {
            w = std::min(v_weights[j],v_weights[spJA[i]]);
            r = spA[i].ne;

            #ifdef CG_DEBUG
            //printf("(%f,%f) ", tc, tn);
            #endif
            
            #ifdef INSPECT_COMPOSITE_GRAPH
            //printf("row %d col %d vrow %d vcol %d w %d r %d\n",
            //spJA[i],j,v_weights[spJA[i]],v_weights[j],w,r);
            #endif
            if (w > 1) {
                if (r>0) {
                    unique_edges++; // r>0 is required b/c (0,0) gets added regardless; idk why
                    conserved_duration(spA[i].nl, spA[i].startl, spA[i].endl,
                                       spA[i].nr, spA[i].startr, spA[i].endr,
                                       &tc, &tn);
                    Tc += tc;
                    Tn += tn;
                }
                if (r>1) cis++; //(float)r / (float)w;
            }
        }
    }
    *n_edges_preserved = cis;
    *n_unique_edges = unique_edges;
    *t_duration_conserved = Tc;
    *t_duration_nonconserved = Tn;
}

void alignment_compute(struct alignment* a, struct carrier* rel,
                       compute_aux_space *caux) {

    alignment_composite_graph(a,caux);
    alignment_edges_preserved(a, caux,&(a->n_edges_preserved), &(a->n_edges_unique),
                              &(a->t_duration_conserved), &(a->t_duration_nonconserved));
    //    alignment_cg_new(a, caux, &(a->n_edges_preserved), &(a->n_edges_unique));
    a->n_edges_induced = a->n_edges_unique + a->n_edges_preserved - a->domain->n_edges;
#ifdef CG_DEBUG
    //printf("conserved_edges: %d\tunique_edges: %d\tconserved_duration: %f\tnonconserved_duration: %f\n----\n", a->n_edges_preserved,a->n_edges_unique,a->t_duration_conserved,a->t_duration_nonconserved);
#endif
    //printf("n_edges_preserved %d n_edges_unique %d s3: %f\n", a->n_edges_preserved,a->n_edges_unique,(float)a->n_edges_preserved/a->n_edges_unique);

//     alignment_composite_graph_new(a,caux);
//     alignment_edges_preserved(a, caux,&(a->n_edges_preserved), &(a->n_edges_unique));
//     printf("s3_score_new %f\n", (float)a->n_edges_preserved/a->n_edges_unique);

    switch (rel->rel) {
    case 0:
        a->edge_score = (float)a->n_edges_preserved/a->domain->n_edges; break;
    case 1:
        a->edge_score = (float)a->n_edges_preserved/a->n_edges_induced; break;
    case 2:
        a->edge_score = (float)a->n_edges_preserved/a->n_edges_unique; break;
    default:
        a->edge_score = (float)a->t_duration_conserved/(a->t_duration_conserved+a->t_duration_nonconserved); break;
    }

    //printf("%f\n", a->edge_score);

    /* add node score here */
    if (rel->use_alpha) {
        // a->node_score = 1.0/(1+alignment_nodescore_compute(a,rel)); /* if distance compare */
        // a->node_score = -nodescore_compute(a,rel); /* if distance compare !! magna cannot handle negative scores though so don't do this */
        a->node_score = alignment_nodescore_compute(a,rel); /* if similarity scores */
        a->score = (a->edge_score)*(rel->alpha) + a->node_score*(1-rel->alpha);
    }
    else {
        a->score = a->edge_score;
        a->node_score = 0.0;
    }

//    printf("alpha: %f, score: %f\n", rel->alpha, a->score);
//    printf("dom: %d, rge: %d, pdeg: %d\n", a->domain->n_vertices,
//           a->range->n_vertices, a->perm->degree);

    a->is_computed = 1;
}

float alignment_nodescore_compute(struct alignment *a, struct carrier *rel) {
    //    int *sequence = a->perm->sequence;
    int n_dom = a->domain->n_vertices;
    int n_ran = a->range->n_vertices;
    float sum = 0;

    int row,col;
    for (row=0; row<n_dom; row++) {
        //        col = sequence[row];
        col = evaluate(a->perm, row);
        sum += rel->cmpdata[row*n_ran+col]; /* row major order */
    }

    return (sum/(float)n_dom);
}

template <class T> void swap ( T& a, T& b )
{
    T c(a); a=b; b=c;
}
 
void alignment_write(struct alignment* a, char* file_name) {
    struct graph* domain = a->domain;
    struct graph* range = a->range;
    int n_vertices = domain->n_vertices;
    int flipgorder = a->flipgorder;
	
    FILE* output_file = NULL;
    if (file_name != NULL) {
        output_file = fopen(file_name, "w");
        if (output_file == NULL) {
            printf("alignment_write: Couldn't open file: %s\n", file_name);
            mg_quit(EXIT_FAILURE);
        }
    }
		
    int i;
    int j;
    char* name1;
    char* name2;
    char *nametemp;
    for (i = 0; i < n_vertices; i++) {
        name1 = graph_find_vertex_name(domain, i);
        j = evaluate(a->perm, i);
        name2 = graph_find_vertex_name(range, j);
        if (flipgorder) {
            nametemp = name1;
            name1 = name2;
            name2 = nametemp;
        }
        if (file_name == NULL)
            printf("%s %s\n", name1, name2);
        else
            fprintf(output_file, "%s %s\n", name1, name2);
    }

    //permutation_print(a->perm);

    if (file_name!=NULL) fclose(output_file);
}

void alignment_common_subgraph_write(struct alignment* a, char* file_name,
                                     compute_aux_space *caux) {
    //alignment_cg_new(a, caux, &(a->n_edges_preserved), &(a->n_edges_unique));
    alignment_composite_graph(a,caux);
    //    SpMat &composite_graph = caux->cg;
    struct compmatelt *spA = caux->spA; // values
    int *spIA = caux->spIA; // ind first elt of each row
    int *spJA = caux->spJA; // col inds
    int *spINZ = caux->spINZ; // nnz of each row
    int *v_weights = caux->v_weights;    
    struct graph* domain = a->domain;
    struct graph* range = a->range;
    int nk = range->n_vertices;
    int i,j,w,r;
    char* name1, *name2;
    int flipgorder = a->flipgorder;

    FILE* output_file = NULL;
    output_file = fopen(file_name, "w");
    if (output_file == NULL) {
        printf("alignment_graph_write: Couldn't open file: %s\n", file_name);
        mg_quit(EXIT_FAILURE);
    }

    struct permutation *q = permutation_calloc(a->perm->degree);
    inverse(q,a->perm);

    for (j=0;j<nk;j++) {
        for(i=spIA[j]; i<spIA[j] + spINZ[j]; i++) {
            w = std::min(v_weights[j],v_weights[spJA[i]]);
            r = spA[i].ne;
            if (w > 1 && r>1) {
                name1 = graph_find_vertex_name(range, j);
                name2 = graph_find_vertex_name(domain, evaluate(q,j));
                if (flipgorder) swap(name1,name2);
                fprintf(output_file, "%s=%s i ", name2, name1);
                name1 = graph_find_vertex_name(range, spJA[i]);
                name2 = graph_find_vertex_name(domain, evaluate(q,spJA[i]));
                if (flipgorder) swap(name1,name2);
                fprintf(output_file, "%s=%s\n", name2, name1);
            }
        }
    }
    permutation_delete(q);
    fclose(output_file);
}

void alignment_graph_write(struct alignment* a, char* file_name) {
    struct graph* domain = a->domain;
    struct graph* range = a->range;
    int flipgorder = a->flipgorder;
	
    FILE* output_file = NULL;
    //	if (file_name != NULL) {
    output_file = fopen(file_name, "w");
    if (output_file == NULL) {
        printf("alignment_graph_write: Couldn't open file: %s\n", file_name);
        mg_quit(EXIT_FAILURE);
    }
    //	}
		
    int i;
    int j;
    char* name1;
    char* name2;
    int v1,v2;

    struct spmat spdom = domain->sp;
    struct spmat spran = range->sp;
    for (v1=0;v1<spdom.nr;v1++) {
        for(i=spdom.IA[v1]; i<spdom.IA[v1] + spdom.INZ[v1]; i++) {
            v2 = spdom.JA[i];
            name1 = graph_find_vertex_name(domain, v1);
            name2 = graph_find_vertex_name(domain, v2);
            if (flipgorder) swap(name1,name2);
            fprintf(output_file, "dom_%s i dom_%s\n", name1, name2);
        }
    }
    for (v1=0;v1<spran.nr;v1++) {
        for(i=spran.IA[v1]; i<spran.IA[v1] + spran.INZ[v1]; i++) {
            v2 = spran.JA[i];
            name1 = graph_find_vertex_name(range, v1);
            name2 = graph_find_vertex_name(range, v2);
            if (flipgorder) swap(name1,name2);
            fprintf(output_file, "ran_%s i ran_%s\n", name1, name2);
        }
    }
	for (i = 0; i < domain->n_vertices; i++) {
		name1 = graph_find_vertex_name(domain, i);
		j = evaluate(a->perm, i);
		name2 = graph_find_vertex_name(range, j);
                if (flipgorder) swap(name1,name2);
                fprintf(output_file, "dom_%s a ran_%s\n", name1, name2);
	}
    //permutation_print(a->perm);
	fclose(output_file);
}

void alignment_print(struct alignment* a) {
	alignment_write(a, NULL);
}

void alignment_delete(struct alignment* a) {
	permutation_delete(a->perm);
    //	permutation_delete(a->invperm);
	free(a);
}

// This returns the fraction of maps that are mapped correctly
// Correct as in wrt vertex name
float alignment_node_correctness(struct alignment *a) {
    //    int *sequence = a->perm->sequence;
    int n_dom = a->domain->n_vertices;
    //    int n_ran = a->range->n_vertices;
    int sum = 0;

    //    printf("alignment correctness\n");

	int i,j;
	char* name1;
	char* name2;
	for (i = 0; i < n_dom; i++) {
		name1 = graph_find_vertex_name(a->domain, i);
		j = evaluate(a->perm, i);
		name2 = graph_find_vertex_name(a->range, j);
        sum += (strcmp(name1,name2)==0); // this looks at the whole name
        //sum += (strncmp(name1,name2,1)==0); // this looks at only the first character
        //        printf("dom(%d) %s rge(%d) %s\n", i, name1, j, name2);
    }
    return ((float)sum/n_dom);
}

inline int max ( int a, int b ) { return a > b ? a : b; }

struct compute_aux_space* compute_aux_space_malloc(struct graph *g1, struct graph *g2) {
    struct compute_aux_space* caux = (struct compute_aux_space*)
        malloc(sizeof(*caux));
    //(caux->cg).resize(g2->n_vertices,g2->n_vertices);
    //(caux->cg_triplets).resize(g1->n_edges+g2->n_edges);
    //if(caux->v_weights==NULL) {mg_error("Allocation error."); mg_quit(EXIT_FAILURE);}
    int nnz = g1->n_edges+g2->n_edges;
    int nk = g2->n_vertices;
    caux->spIA = (int*)malloc((nk+1)*sizeof(int)); // ind first elt of each row
    caux->spJA = (int*)malloc(nnz*sizeof(int)); // final col inds
    caux->spJR = (int*)malloc(nnz*sizeof(int)); // row inds
    caux->spJC = (int*)malloc(nnz*sizeof(int)); // col inds
    caux->spINZ = (int*)malloc(nk*sizeof(int)); // nnz of each row
    caux->spWI = (int*)malloc(nk*sizeof(int)); // for the accum. could be done w/ spINZ
    caux->v_weights = (int*)malloc(nk*sizeof(int)); // vector of "cluster sizes"

    int i;
    caux->tempA = (struct compmatelt*)malloc(nnz*sizeof(struct compmatelt)); // number of events
    //caux->tempAalloc = (int*)malloc(nnz*sizeof(int)); // number of events allocated
    //caux->tempAstart = (float**)malloc(nnz*sizeof(float*));
    //caux->tempAend = (float**)malloc(nnz*sizeof(float*));

    int n1=0,n2=0;
    for(i=0;i<g1->n_edges;i++) n1 = max(n1,g1->sp.A[i].n);
    for(i=0;i<g2->n_edges;i++) n2 = max(n2,g2->sp.A[i].n);
    int ntot = 4; //n1+n2;
    
    //printf("HH %d ",(ntot*sizeof(float)));
    //for(i=0;i<nnz;i++) printf("FF(%d,%d,%d) ", malloc_usable_size(caux->tempAstart[i]), caux->tempAalloc[i], caux->tempA[i]);

    caux->spA = (struct compmatelt*)malloc(nnz*sizeof(struct compmatelt)); // number of events
    //caux->spAalloc = (int*)malloc(nnz*sizeof(int)); // number of events allocated
    //caux->spAstart = (float**)malloc(nnz*sizeof(float*));
    //caux->spAend = (float**)malloc(nnz*sizeof(float*));
    //printf("Done allocating auxilary space\n");
    caux->tempstart = (float*)malloc((n1+n2)*sizeof(float));
    caux->tempend = (float*)malloc((n1+n2)*sizeof(float));

    if (caux->v_weights==NULL || caux->spIA==NULL || caux->spJA==NULL ||
        caux->spJR==NULL || caux->spJC==NULL ||caux->spINZ==NULL ||caux->spWI==NULL ||
        caux->tempA==NULL || caux->spA==NULL || caux->tempstart==NULL || caux->tempend==NULL)
        {mg_error("Allocation error."); mg_quit(EXIT_FAILURE);}

    caux->nnz = nnz;
    
    return(caux);
}

void compute_aux_space_delete(struct compute_aux_space* caux) {
    int nnz = caux->nnz;
    int i;
    //for (i=0;i<nnz;i++) free(caux->spA[i].start);
    //for (i=0;i<nnz;i++) free(caux->spA[i].end);
    free(caux->tempstart);
    free(caux->tempend);
    free(caux->spA);
    free(caux->tempA);
    free(caux->spIA);
    free(caux->spJA);
    free(caux->spJR);
    free(caux->spJC);
    free(caux->spINZ);
    free(caux->spWI);
    free(caux->v_weights);
    free(caux);
}
